/*
 * Shared Memory definitions & API
 *
 * Copyright (C) 2013-2016 CERN (www.cern.ch)
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __WRNODE_SMEM_H
#define __WRNODE_SMEM_H


/**
 * List of available access modes
 */
#define SMEM_RANGE_ADD		0x10000
#define SMEM_RANGE_SUB		0x20000
#define SMEM_RANGE_SET		0x30000
#define SMEM_RANGE_CLEAR	0x40000
#define SMEM_RANGE_FLIP		0x50000
#define SMEM_RANGE_TEST_AND_SET	0x60000

#define SMEM volatile __attribute__((section(".smem")))

/**
 * Perform an operation on a given pointer. Operation can be performed only
 * on integer variables
 */
static inline void __smem_atomic_op(volatile int *p, int x, unsigned int range)
{
	*(volatile int *)(p + (range >> 2)) = x;
}


/**
 * Add x to the value in the shared memory pointed by p
 */
static inline void smem_atomic_add(volatile int *p, int x)
{
	__smem_atomic_op(p, x, SMEM_RANGE_ADD);
}


/**
 * Subtract x to the value in the shared memory pointed by p
 */
static inline void smem_atomic_sub(volatile int *p, int x)
{
	__smem_atomic_op(p, x, SMEM_RANGE_SUB);
}


/**
 * Do "OR" between x and the value in the shared memory pointed by p
 */
static inline void smem_atomic_or(volatile int *p, int x)
{
	__smem_atomic_op(p, x, SMEM_RANGE_SET);
}


/**
 * Do "AND ~" between x and the value in the shared memory pointed by p
 */
static inline void smem_atomic_and_not(volatile int *p, int x)
{
	__smem_atomic_op(p, x, SMEM_RANGE_CLEAR);
}


/**
 * Do "XOR" between x and the value in the shared memory pointed by p
 */
static inline void smem_atomic_xor(int *p, int x)
{
	__smem_atomic_op(p, x, SMEM_RANGE_FLIP);
}

/**
 * Do test & set in the shared memory pointed by p
 * The current value is returned and the new value is setted to 1
 * This allows to implement mutex
 */
static inline int smem_atomic_test_and_set(int *p)
{
	/* shift right translates range in bytes into range in words */
	return *(volatile int *)(p + (SMEM_RANGE_TEST_AND_SET >> 2));
}

#endif
